﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HEAL.Attic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;

namespace HeuristicLab.Problems.Modifiers {
  [StorableType("03EC0CED-F927-4CCF-8262-716AECE267BE")]
  [Item("MultiObjectiveProblemModifier", "A modifier attached to a problem")]
  public abstract class ProblemModifier : ParameterizedNamedItem {

    [Storable] private ProblemModifier nextModifier;
    public virtual IEncoding Encoding => NextModifier?.Encoding;
    public virtual BoolArray Maximization => NextModifier?.Maximization;
    public virtual ISolutionCreator SolutionCreator => NextModifier?.SolutionCreator;
    public ProblemModifier NextModifier {
      get => nextModifier;
      set
      {
        if (nextModifier == value) return;
        ResetListeners(nextModifier, value);
        nextModifier = value;
        InvokeEncodingChanged();
        InvokeMaximizationChanged();
      }
    }

    #region Constructors
    [StorableConstructor]
    protected ProblemModifier(StorableConstructorFlag _) : base(_) { }
    protected ProblemModifier(ProblemModifier original, Cloner cloner) : base(original, cloner) {
      nextModifier = cloner.Clone(original.NextModifier);
    }
    protected ProblemModifier() { }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      ResetListeners(nextModifier, nextModifier);
    }
    #endregion

    public virtual void Initialize() { }

    public virtual double[] ModifiedEvaluate(Individual individual, IRandom random) {
      return nextModifier.ModifiedEvaluate(individual, random);
    }

    public virtual void ModifiedAnalyze(Individual[] individuals, double[][] qualities, ResultCollection results, IRandom random) {
      nextModifier.ModifiedAnalyze(individuals, qualities, results, random);
    }

    #region updateEvents
    public event EventHandler EncodingChanged;
    public event EventHandler MaximizationChanged;
    public event EventHandler SolutionCreatorChanged;
    protected void InvokeEncodingChanged() {
      EncodingChanged?.Invoke(this, EventArgs.Empty);
    }
    protected void InvokeMaximizationChanged() {
      MaximizationChanged?.Invoke(this, EventArgs.Empty);
    }
    protected void InvokeSolutionCreatorChanged() {
      SolutionCreatorChanged?.Invoke(this, EventArgs.Empty);
    }
    protected virtual void OnNextMaximizationChanged(object sender, EventArgs e) {
      InvokeMaximizationChanged();
    }
    protected virtual void OnNextEncodingChanged(object sender, EventArgs e) {
      InvokeEncodingChanged();
    }
    protected virtual void OnNextSolutionCreatorChanged(object sender, EventArgs e) {
      InvokeSolutionCreatorChanged();
    }
    #endregion

    #region helpers
    protected virtual void ResetListeners(ProblemModifier oldNext, ProblemModifier newNext) {
      if (oldNext != null) {
        oldNext.EncodingChanged -= OnNextEncodingChanged;
        oldNext.MaximizationChanged -= OnNextMaximizationChanged;
        oldNext.SolutionCreatorChanged -= OnNextSolutionCreatorChanged;
      }
      if (newNext == null) return;
      newNext.EncodingChanged += OnNextEncodingChanged;
      newNext.MaximizationChanged += OnNextMaximizationChanged;
      newNext.SolutionCreatorChanged += OnNextSolutionCreatorChanged;
    }
    #endregion


  }
}
